package java.util.concurrent.atomic;

import java.lang.reflect.Modifier;
import java.security.PrivilegedActionException;
import sun.reflect.misc.ReflectUtil;
import java.security.AccessController;
import java.security.PrivilegedExceptionAction;
import java.lang.reflect.Field;
import sun.misc.Unsafe;
import java.util.Objects;
import java.util.function.LongBinaryOperator;
import java.util.function.LongUnaryOperator;
import sun.reflect.CallerSensitive;
import sun.reflect.Reflection;

public abstract class AtomicLongFieldUpdater<T>
{
    @CallerSensitive
    public static <U> AtomicLongFieldUpdater<U> newUpdater(Class<U> clazz, String s) {
        Class<Object> callerClass = Reflection.getCallerClass();
        if (AtomicLong.VM_SUPPORTS_LONG_CAS) {
            return new AtomicLongFieldUpdater.CASUpdater(clazz, s, callerClass);
        }
        return new AtomicLongFieldUpdater.LockedUpdater(clazz, s, callerClass);
    }
    
    protected AtomicLongFieldUpdater() {
    }
    
    public abstract boolean compareAndSet(T p0, long p1, long p2);
    
    public abstract boolean weakCompareAndSet(T p0, long p1, long p2);
    
    public abstract void set(T p0, long p1);
    
    public abstract void lazySet(T p0, long p1);
    
    public abstract long get(T p0);
    
    public long getAndSet(T t, long n) {
        long value;
        do {
            value = this.get(t);
        } while (!this.compareAndSet(t, value, n));
        return value;
    }
    
    public long getAndIncrement(T t) {
        long value;
        do {
            value = this.get(t);
        } while (!this.compareAndSet(t, value, value + 1L));
        return value;
    }
    
    public long getAndDecrement(T t) {
        long value;
        do {
            value = this.get(t);
        } while (!this.compareAndSet(t, value, value - 1L));
        return value;
    }
    
    public long getAndAdd(T t, long n) {
        long value;
        do {
            value = this.get(t);
        } while (!this.compareAndSet(t, value, value + n));
        return value;
    }
    
    public long incrementAndGet(T t) {
        long value;
        long n;
        do {
            value = this.get(t);
            n = value + 1L;
        } while (!this.compareAndSet(t, value, n));
        return n;
    }
    
    public long decrementAndGet(T t) {
        long value;
        long n;
        do {
            value = this.get(t);
            n = value - 1L;
        } while (!this.compareAndSet(t, value, n));
        return n;
    }
    
    public long addAndGet(T t, long n) {
        long value;
        long n2;
        do {
            value = this.get(t);
            n2 = value + n;
        } while (!this.compareAndSet(t, value, n2));
        return n2;
    }
    
    public final long getAndUpdate(T t, LongUnaryOperator longUnaryOperator) {
        long value;
        do {
            value = this.get(t);
        } while (!this.compareAndSet(t, value, longUnaryOperator.applyAsLong(value)));
        return value;
    }
    
    public final long updateAndGet(T t, LongUnaryOperator longUnaryOperator) {
        long value;
        long applyAsLong;
        do {
            value = this.get(t);
            applyAsLong = longUnaryOperator.applyAsLong(value);
        } while (!this.compareAndSet(t, value, applyAsLong));
        return applyAsLong;
    }
    
    public final long getAndAccumulate(T t, long n, LongBinaryOperator longBinaryOperator) {
        long value;
        do {
            value = this.get(t);
        } while (!this.compareAndSet(t, value, longBinaryOperator.applyAsLong(value, n)));
        return value;
    }
    
    public final long accumulateAndGet(T t, long n, LongBinaryOperator longBinaryOperator) {
        long value;
        long applyAsLong;
        do {
            value = this.get(t);
            applyAsLong = longBinaryOperator.applyAsLong(value, n);
        } while (!this.compareAndSet(t, value, applyAsLong));
        return applyAsLong;
    }
    
    static boolean isAncestor(ClassLoader parent, ClassLoader classLoader) {
        do {
            parent = parent.getParent();
            if (classLoader == parent) {
                return true;
            }
            continue;
        } while (parent != null);
        return false;
    }
    
    private static boolean isSamePackage(Class<?> clazz, Class<?> clazz2) {
        return clazz.getClassLoader() == clazz2.getClassLoader() && Objects.equals(getPackageName(clazz), getPackageName(clazz2));
    }
    
    private static String getPackageName(Class<?> clazz) {
        String name = clazz.getName();
        int lastIndex = name.lastIndexOf(46);
        return (lastIndex != -1) ? name.substring(0, lastIndex) : "";
    }
    
    static boolean access$000(Class clazz, Class clazz2) {
        return isSamePackage(clazz, clazz2);
    }
    
    private static final class CASUpdater<T> extends AtomicLongFieldUpdater<T>
    {
        private static final Unsafe U;
        private final long offset;
        private final Class<?> cclass;
        private final Class<T> tclass;
        
        CASUpdater(Class<T> tclass, String s, Class<?> currentClass) {
            Field f;
            int modifiers;
            try {
                f = (Field)AccessController.doPrivileged(new PrivilegedExceptionAction<Field>() {
                    @Override
                    public Field run() throws NoSuchFieldException {
                        return tclass.getDeclaredField(s);
                    }
                });
                modifiers = f.getModifiers();
                ReflectUtil.ensureMemberAccess(currentClass, tclass, null, modifiers);
                ClassLoader classLoader = tclass.getClassLoader();
                ClassLoader classLoader2 = currentClass.getClassLoader();
                if (classLoader2 != null && classLoader2 != classLoader && (classLoader == null || !AtomicLongFieldUpdater.isAncestor(classLoader, classLoader2))) {
                    ReflectUtil.checkPackageAccess(tclass);
                }
            }
            catch (PrivilegedActionException ex) {
                throw new RuntimeException(ex.getException());
            }
            catch (Exception cause) {
                throw new RuntimeException(cause);
            }
            if (f.getType() != Long.TYPE) {
                throw new IllegalArgumentException("Must be long type");
            }
            if (!Modifier.isVolatile(modifiers)) {
                throw new IllegalArgumentException("Must be volatile type");
            }
            this.cclass = ((Modifier.isProtected(modifiers) && tclass.isAssignableFrom(currentClass) && !AtomicLongFieldUpdater.access$000(tclass, currentClass)) ? currentClass : tclass);
            this.tclass = tclass;
            this.offset = AtomicLongFieldUpdater.CASUpdater.U.objectFieldOffset(f);
            return;
        }
        
        private final void accessCheck(T t) {
            if (!this.cclass.isInstance(t)) {
                this.throwAccessCheckException(t);
            }
        }
        
        private final void throwAccessCheckException(T t) {
            if (this.cclass == this.tclass) {
                throw new ClassCastException();
            }
            throw new RuntimeException(new IllegalAccessException("Class " + this.cclass.getName() + " can not access a protected member of class " + this.tclass.getName() + " using an instance of " + t.getClass().getName()));
        }
        
        @Override
        public final boolean compareAndSet(T o, long expected, long x) {
            this.accessCheck(o);
            return AtomicLongFieldUpdater.CASUpdater.U.compareAndSwapLong(o, this.offset, expected, x);
        }
        
        @Override
        public final boolean weakCompareAndSet(T o, long expected, long x) {
            this.accessCheck(o);
            return AtomicLongFieldUpdater.CASUpdater.U.compareAndSwapLong(o, this.offset, expected, x);
        }
        
        @Override
        public final void set(T o, long x) {
            this.accessCheck(o);
            AtomicLongFieldUpdater.CASUpdater.U.putLongVolatile(o, this.offset, x);
        }
        
        @Override
        public final void lazySet(T o, long x) {
            this.accessCheck(o);
            AtomicLongFieldUpdater.CASUpdater.U.putOrderedLong(o, this.offset, x);
        }
        
        @Override
        public final long get(T o) {
            this.accessCheck(o);
            return AtomicLongFieldUpdater.CASUpdater.U.getLongVolatile(o, this.offset);
        }
        
        @Override
        public final long getAndSet(T o, long newValue) {
            this.accessCheck(o);
            return AtomicLongFieldUpdater.CASUpdater.U.getAndSetLong(o, this.offset, newValue);
        }
        
        @Override
        public final long getAndAdd(T o, long delta) {
            this.accessCheck(o);
            return AtomicLongFieldUpdater.CASUpdater.U.getAndAddLong(o, this.offset, delta);
        }
        
        @Override
        public final long getAndIncrement(T t) {
            return this.getAndAdd(t, 1L);
        }
        
        @Override
        public final long getAndDecrement(T t) {
            return this.getAndAdd(t, -1L);
        }
        
        @Override
        public final long incrementAndGet(T t) {
            return this.getAndAdd(t, 1L) + 1L;
        }
        
        @Override
        public final long decrementAndGet(T t) {
            return this.getAndAdd(t, -1L) - 1L;
        }
        
        @Override
        public final long addAndGet(T t, long n) {
            return this.getAndAdd(t, n) + n;
        }
        
        static {
            U = Unsafe.getUnsafe();
        }
    }
    
    private static final class LockedUpdater<T> extends AtomicLongFieldUpdater<T>
    {
        private static final Unsafe U;
        private final long offset;
        private final Class<?> cclass;
        private final Class<T> tclass;
        
        LockedUpdater(Class<T> tclass, String s, Class<?> currentClass) {
            Field f;
            int modifiers;
            try {
                f = (Field)AccessController.doPrivileged(new PrivilegedExceptionAction<Field>() {
                    @Override
                    public Field run() throws NoSuchFieldException {
                        return tclass.getDeclaredField(s);
                    }
                });
                modifiers = f.getModifiers();
                ReflectUtil.ensureMemberAccess(currentClass, tclass, null, modifiers);
                ClassLoader classLoader = tclass.getClassLoader();
                ClassLoader classLoader2 = currentClass.getClassLoader();
                if (classLoader2 != null && classLoader2 != classLoader && (classLoader == null || !AtomicLongFieldUpdater.isAncestor(classLoader, classLoader2))) {
                    ReflectUtil.checkPackageAccess(tclass);
                }
            }
            catch (PrivilegedActionException ex) {
                throw new RuntimeException(ex.getException());
            }
            catch (Exception cause) {
                throw new RuntimeException(cause);
            }
            if (f.getType() != Long.TYPE) {
                throw new IllegalArgumentException("Must be long type");
            }
            if (!Modifier.isVolatile(modifiers)) {
                throw new IllegalArgumentException("Must be volatile type");
            }
            this.cclass = ((Modifier.isProtected(modifiers) && tclass.isAssignableFrom(currentClass) && !AtomicLongFieldUpdater.access$000(tclass, currentClass)) ? currentClass : tclass);
            this.tclass = tclass;
            this.offset = AtomicLongFieldUpdater.LockedUpdater.U.objectFieldOffset(f);
            return;
        }
        
        private final void accessCheck(T t) {
            if (!this.cclass.isInstance(t)) {
                throw this.accessCheckException(t);
            }
            return;
        }
        
        private final RuntimeException accessCheckException(T t) {
            if (this.cclass == this.tclass) {
                return new ClassCastException();
            }
            return new RuntimeException(new IllegalAccessException("Class " + this.cclass.getName() + " can not access a protected member of class " + this.tclass.getName() + " using an instance of " + t.getClass().getName()));
        }
        
        @Override
        public final boolean compareAndSet(T t, long n, long x) {
            this.accessCheck(t);
            synchronized (this) {
                if (AtomicLongFieldUpdater.LockedUpdater.U.getLong(t, this.offset) != n) {
                    return false;
                }
                AtomicLongFieldUpdater.LockedUpdater.U.putLong(t, this.offset, x);
                return true;
            }
        }
        
        @Override
        public final boolean weakCompareAndSet(T t, long n, long n2) {
            return this.compareAndSet(t, n, n2);
        }
        
        @Override
        public final void set(T o, long x) {
            this.accessCheck(o);
            synchronized (this) {
                AtomicLongFieldUpdater.LockedUpdater.U.putLong(o, this.offset, x);
            }
        }
        
        @Override
        public final void lazySet(T t, long n) {
            this.set(t, n);
        }
        
        @Override
        public final long get(T o) {
            this.accessCheck(o);
            synchronized (this) {
                return AtomicLongFieldUpdater.LockedUpdater.U.getLong(o, this.offset);
            }
        }
        
        static {
            U = Unsafe.getUnsafe();
        }
    }
}
